@@ -2,9 +2,11 @@ from typing import TYPE_CHECKING
 
 if TYPE_CHECKING:
     from policyengine_core.populations import Population
-from policyengine_core.projectors import Projector
+
 from numpy.typing import ArrayLike
 
+from policyengine_core.projectors.projector import Projector
+
 
 class EntityToPersonProjector(Projector):
     """For instance person.family."""
