@@ -1,14 +1,16 @@
 import tempfile
-from typing import Any, Dict, List, TYPE_CHECKING, Type
+from typing import TYPE_CHECKING, Any, Dict, List, Type, Union
+
 import numpy
 import numpy as np
 import pandas as pd
 from numpy.typing import ArrayLike
+
 from policyengine_core import commons, periods
 from policyengine_core.data.dataset import Dataset
 from policyengine_core.entities.entity import Entity
-from policyengine_core.errors import CycleError, SpiralError
 from policyengine_core.enums import Enum, EnumArray
+from policyengine_core.errors import CycleError, SpiralError
 from policyengine_core.holders.holder import Holder
 from policyengine_core.periods import Period
 from policyengine_core.periods.config import ETERNITY
@@ -21,10 +23,13 @@ from policyengine_core.tracers import (
 
 if TYPE_CHECKING:
     from policyengine_core.taxbenefitsystems import TaxBenefitSystem
+
+from policyengine_core.experimental import MemoryConfig
 from policyengine_core.populations import Population
 from policyengine_core.tracers import SimpleTracer
-from policyengine_core.experimental import MemoryConfig
 from policyengine_core.variables import Variable
+from policyengine_core.reforms.reform import Reform
+from policyengine_core.parameters import get_parameter
 
 
 class Simulation:
@@ -35,6 +40,9 @@ class Simulation:
     default_tax_benefit_system: Type["TaxBenefitSystem"] = None
     """The default tax-benefit system class to use if none is provided."""
 
+    default_tax_benefit_system_instance: "TaxBenefitSystem" = None
+    """The default tax-benefit system instance to use if none is provided. This requires that the tax-benefit system is initialised when importing a country package. This will slow down the import, but may speed up individual simulations."""
+
     default_dataset: Dataset = None
     """The default dataset class to use if none is provided."""
 
@@ -44,6 +52,12 @@ class Simulation:
     default_role: str = None
     """The default role to assign people to groups if none is provided."""
 
+    default_input_period: str = None
+    """The default period to use when inputting variables."""
+
+    default_calculation_period: str = None
+    """The default period to calculate for if none is provided."""
+
     def __init__(
         self,
         tax_benefit_system: "TaxBenefitSystem" = None,
@@ -51,16 +65,25 @@ class Simulation:
         situation: dict = None,
         dataset: Type[Dataset] = None,
         dataset_year: int = None,
+        reform: Reform = None,
     ):
-        """
-        This constructor is reserved for internal use; see :any:`SimulationBuilder`,
-        which is the preferred way to obtain a Simulation initialized with a consistent
-        set of Entities.
-        """
         if tax_benefit_system is None:
-            tax_benefit_system = self.default_tax_benefit_system()
+            if self.default_tax_benefit_system_instance is not None:
+                tax_benefit_system = self.default_tax_benefit_system_instance
+                if reform is not None:
+                    tax_benefit_system = tax_benefit_system.clone()
+            else:
+                tax_benefit_system = self.default_tax_benefit_system()
+            self.tax_benefit_system = tax_benefit_system
+
+        self.reform = reform
         self.tax_benefit_system = tax_benefit_system
 
+        if reform is not None:
+            self.apply_reform(reform)
+
+        self.branch_name = "default"
+
         if dataset is None:
             if self.default_dataset is not None:
                 dataset = self.default_dataset
@@ -79,6 +102,9 @@ class Simulation:
         self.memory_config: MemoryConfig = None
         self._data_storage_dir: str = None
 
+        self.branches: Dict[str, Simulation] = {}
+        self.has_axes = False
+
         if situation is not None:
             if dataset is not None:
                 raise ValueError(
@@ -91,9 +117,10 @@ class Simulation:
                 SimulationBuilder,
             )  # Import here to avoid circular dependency
 
-            SimulationBuilder().build_from_dict(
-                self.tax_benefit_system, situation, self
-            )
+            builder = SimulationBuilder()
+            builder.default_period = self.default_input_period
+            builder.build_from_dict(self.tax_benefit_system, situation, self)
+            self.has_axes = builder.has_axes
 
         if populations is not None:
             self.build_from_populations(populations)
@@ -103,6 +130,23 @@ class Simulation:
             self.dataset_year = dataset_year
             self.build_from_dataset()
 
+        # Backwards compatibility methods
+        self.calc = self.calculate
+        self.df = self.calculate_dataframe
+
+        self.input_variables = [
+            variable.name
+            for variable in self.tax_benefit_system.variables.values()
+            if len(self.get_holder(variable.name).get_known_periods()) > 0
+        ]
+
+    def apply_reform(self, reform: Union[tuple, Reform]):
+        if isinstance(reform, tuple):
+            for subreform in reform:
+                self.apply_reform(subreform)
+        else:
+            reform.apply(self.tax_benefit_system)
+
     def build_from_populations(
         self, populations: Dict[str, Population]
     ) -> None:
@@ -148,7 +192,7 @@ class Simulation:
         ), f"Missing {entity_id_field} column in the dataset. Each person entity must have an ID array defined for ETERNITY."
 
         get_eternity_array = (
-            lambda ds: ds[ETERNITY]
+            lambda ds: ds[list(ds.keys())[0]]
             if self.dataset.data_format == Dataset.TIME_PERIOD_ARRAYS
             else ds
         )
@@ -164,10 +208,15 @@ class Simulation:
             entity_ids = get_eternity_array(data[entity_id_field])
             builder.declare_entity(group_entity.key, entity_ids)
 
-            person_membership_id = f"{person_entity.key}_{group_entity.key}_id"
+            person_membership_id_field = (
+                f"{person_entity.key}_{group_entity.key}_id"
+            )
             assert (
-                person_membership_id in data
-            ), f"Missing {person_membership_id} column in the dataset. Each group entity must have a person membership array defined for ETERNITY."
+                person_membership_id_field in data
+            ), f"Missing {person_membership_id_field} column in the dataset. Each group entity must have a person membership array defined for ETERNITY."
+            person_membership_ids = get_eternity_array(
+                data[person_membership_id_field]
+            )
 
             person_role_field = f"{person_entity.key}_{group_entity.key}_role"
             if person_role_field in data:
@@ -182,10 +231,12 @@ class Simulation:
                 )
             builder.join_with_persons(
                 self.populations[group_entity.key],
-                person_membership_id,
+                person_membership_ids,
                 person_roles,
             )
 
+        self.build_from_populations(builder.populations)
+
         for variable in data:
             if variable in self.tax_benefit_system.variables:
                 if self.dataset.data_format == Dataset.TIME_PERIOD_ARRAYS:
@@ -238,7 +289,10 @@ class Simulation:
     # ----- Calculation methods ----- #
 
     def calculate(
-        self, variable_name: str, period: Period, map_to: str = None
+        self,
+        variable_name: str,
+        period: Period = None,
+        map_to: str = None,
     ) -> ArrayLike:
         """Calculate ``variable_name`` for ``period``.
 
@@ -253,8 +307,12 @@ class Simulation:
 
         if period is not None and not isinstance(period, Period):
             period = periods.period(period)
+        elif period is None and self.default_calculation_period is not None:
+            period = periods.period(self.default_calculation_period)
 
-        self.tracer.record_calculation_start(variable_name, period)
+        self.tracer.record_calculation_start(
+            variable_name, period, self.branch_name
+        )
 
         try:
             result = self._calculate(variable_name, period)
@@ -265,7 +323,6 @@ class Simulation:
                 ).entity.key
                 result = self.map_result(result, source_entity, map_to)
             return result
-
         finally:
             self.tracer.record_calculation_end()
             self.purge_cache_of_invalid_values()
@@ -295,7 +352,7 @@ class Simulation:
         target_pop = self.populations[target_entity]
         if (
             source_entity == "person"
-            and target_entity in self.group_entity_names
+            and target_entity in self.tax_benefit_system.group_entity_keys
         ):
             if how and how not in (
                 "sum",
@@ -308,7 +365,7 @@ class Simulation:
                 raise ValueError("Not a valid function.")
             return target_pop.__getattribute__(how or "sum")(values)
         elif (
-            source_entity in self.group_entity_names
+            source_entity in self.tax_benefit_system.group_entity_keys
             and target_entity == "person"
         ):
             if not how:
@@ -318,8 +375,8 @@ class Simulation:
         elif source_entity == target_entity:
             return values
         else:
-            return self.map_to(
-                self.map_to(
+            return self.map_result(
+                self.map_result(
                     values,
                     source_entity,
                     self.tax_benefit_system.person_entity.key,
@@ -331,7 +388,10 @@ class Simulation:
             )
 
     def calculate_dataframe(
-        self, variable_names: List[str], period: Period, map_to: str = None
+        self,
+        variable_names: List[str],
+        period: Period = None,
+        map_to: str = None,
     ) -> pd.DataFrame:
         """Calculate ``variable_names`` for ``period``.
 
@@ -356,7 +416,9 @@ class Simulation:
             df[variable_name] = self.calculate(variable_name, period, map_to)
         return df
 
-    def _calculate(self, variable_name: str, period: Period) -> ArrayLike:
+    def _calculate(
+        self, variable_name: str, period: Period = None
+    ) -> ArrayLike:
         """
         Calculate the variable ``variable_name`` for the period ``period``, using the variable formula if it exists.
 
@@ -376,10 +438,23 @@ class Simulation:
         self._check_period_consistency(period, variable)
 
         # First look for a value already cached
-        cached_array = holder.get_array(period)
+        cached_array = holder.get_array(period, self.branch_name)
         if cached_array is not None:
             return cached_array
 
+        if variable.defined_for is not None:
+            mask = (
+                self.calculate(
+                    variable.defined_for, period, map_to=variable.entity.key
+                )
+                > 0
+            )
+            if np.all(~mask):
+                array = holder.default_array()
+                array = self._cast_formula_result(array, variable)
+                holder.put_in_cache(array, period, self.branch_name)
+                return array
+
         array = None
 
         # First, try to run a formula
@@ -389,13 +464,67 @@ class Simulation:
 
             # If no result, use the default value and cache it
             if array is None:
-                array = holder.default_array()
+                # Check if the variable has a previously defined value
+                known_periods = holder.get_known_periods()
+                if variable.uprating is not None and len(known_periods) > 0:
+                    start_instants = [
+                        known_period.start for known_period in known_periods
+                    ]
+                    latest_known_period = known_periods[
+                        np.argmax(start_instants)
+                    ]
+                    if latest_known_period.start < period.start:
+                        try:
+                            uprating_parameter = get_parameter(
+                                self.tax_benefit_system.parameters,
+                                variable.uprating,
+                            )
+                        except:
+                            raise ValueError(
+                                f"Could not find uprating parameter {variable.uprating} when trying to uprate {variable_name}."
+                            )
+                        value_in_last_period = uprating_parameter(
+                            latest_known_period.start
+                        )
+                        value_in_this_period = uprating_parameter(period.start)
+                        uprating_factor = (
+                            value_in_this_period / value_in_last_period
+                        )
+                        return (
+                            holder.get_array(
+                                latest_known_period, self.branch_name
+                            )
+                            * uprating_factor
+                        )
+                elif (
+                    self.tax_benefit_system.auto_carry_over_input_variables
+                    and variable.calculate_output is None
+                    and len(known_periods) > 0
+                ):
+                    # Variables with a calculate-output property specify
+                    last_known_period = sorted(known_periods)[-1]
+                    array = holder.get_array(last_known_period)
+                else:
+                    array = holder.default_array()
+
+            if variable.defined_for is not None:
+                array = np.where(mask, array, np.zeros_like(array))
 
             array = self._cast_formula_result(array, variable)
-            holder.put_in_cache(array, period)
+            holder.put_in_cache(array, period, self.branch_name)
 
         except SpiralError:
             array = holder.default_array()
+        except RecursionError:
+            if self.trace:
+                self.tracer.print_computation_log()
+            else:
+                print(
+                    "Computation log is only available with the trace=True option."
+                )
+            raise Exception(
+                f"RecursionError while calculating {variable_name} for period {period}. The full computation trace is printed above."
+            )
 
         return array
 
@@ -408,7 +537,9 @@ class Simulation:
             holder.delete_arrays(_period)
         self.invalidated_caches = set()
 
-    def calculate_add(self, variable_name: str, period: Period) -> ArrayLike:
+    def calculate_add(
+        self, variable_name: str, period: Period = None
+    ) -> ArrayLike:
         variable = self.tax_benefit_system.get_variable(
             variable_name, check_existence=True
         )
@@ -443,7 +574,7 @@ class Simulation:
         )
 
     def calculate_divide(
-        self, variable_name: str, period: Period
+        self, variable_name: str, period: Period = None
     ) -> ArrayLike:
         variable = self.tax_benefit_system.get_variable(
             variable_name, check_existence=True
@@ -480,7 +611,7 @@ class Simulation:
         )
 
     def calculate_output(
-        self, variable_name: str, period: Period
+        self, variable_name: str, period: Period = None
     ) -> ArrayLike:
         """
         Calculate the value of a variable using the ``calculate_output`` attribute of the variable.
@@ -499,6 +630,7 @@ class Simulation:
         return TracingParameterNodeAtInstant(
             self.tax_benefit_system.get_parameters_at_instant(formula_period),
             self.tracer,
+            self.branch_name,
         )
 
     def _run_formula(
@@ -510,7 +642,21 @@ class Simulation:
 
         formula = variable.get_formula(period)
         if formula is None:
-            return None
+            values = None
+            if variable.adds is not None:
+                values = 0
+                for added_variable in variable.adds:
+                    values = values + self.calculate(
+                        added_variable, period, map_to=variable.entity.key
+                    )
+            if variable.subtracts is not None:
+                if values is None:
+                    values = 0
+                for subtracted_variable in variable.subtracts:
+                    values = values - self.calculate(
+                        subtracted_variable, period, map_to=variable.entity.key
+                    )
+            return values
 
         if self.trace:
             parameters_at = self.trace_parameters_at_instant
@@ -591,11 +737,29 @@ class Simulation:
             frame["period"]
             for frame in self.tracer.stack[:-1]
             if frame["name"] == variable
+            and frame["branch_name"] == self.branch_name
         ]
         if period in previous_periods:
+            found_last_frame = False
+            i = -2
+            while not found_last_frame:
+                frame = self.tracer.stack[i]
+                if (
+                    frame["name"] == variable
+                    and frame["branch_name"] == self.branch_name
+                ):
+                    found_last_frame = True
+                i -= 1
             raise CycleError(
-                "Circular definition detected on formula {}@{}".format(
-                    variable, period
+                f"Circular definition detected on formula {variable}@{period}. The circle is:\n\nNormal computation tree:\n"
+                + "\n".join(
+                    f"  {frame['name']}@{frame['period']} (branch {frame['branch_name']})"
+                    for frame in self.tracer.stack[:i]
+                )
+                + "\n\nCycle start:\n"
+                + "\n".join(
+                    f"  >> {frame['name']}@{frame['period']} (branch {frame['branch_name']})"
+                    for frame in self.tracer.stack[i:]
                 )
             )
         spiral = len(previous_periods) >= self.max_spiral_loops
@@ -632,7 +796,9 @@ class Simulation:
         """
         if period is not None and not isinstance(period, Period):
             period = periods.period(period)
-        return self.get_holder(variable_name).get_array(period)
+        return self.get_holder(variable_name).get_array(
+            period, self.branch_name
+        )
 
     def get_holder(self, variable_name: str) -> Holder:
         """
@@ -726,7 +892,9 @@ class Simulation:
         period = periods.period(period)
         if (variable.end is not None) and (period.start.date > variable.end):
             return
-        self.get_holder(variable_name).set_input(period, value)
+        self.get_holder(variable_name).set_input(
+            period, value, self.branch_name
+        )
 
     def get_variable_population(self, variable_name: str) -> Population:
         variable = self.tax_benefit_system.get_variable(
@@ -762,21 +930,74 @@ class Simulation:
         new_dict = new.__dict__
 
         for key, value in self.__dict__.items():
-            if key not in ("debug", "trace", "tracer"):
+            if key not in ("debug", "trace", "tracer", "branches"):
                 new_dict[key] = value
 
         new.persons = self.persons.clone(new)
         setattr(new, new.persons.entity.key, new.persons)
         new.populations = {new.persons.entity.key: new.persons}
+        new.branches = {}
 
         for entity in self.tax_benefit_system.group_entities:
-            population = self.populations[entity.key].clone(new)
+            population = self.populations[entity.key].clone(new, new.persons)
             new.populations[entity.key] = population
             setattr(
                 new, entity.key, population
             )  # create shortcut simulation.household (for instance)
 
+        new.tax_benefit_system = self.tax_benefit_system.clone()
         new.debug = debug
         new.trace = trace
 
         return new
+
+    def get_branch(self, name: str = "branch") -> "Simulation":
+        """Create a clone of this simulation, whose calculations are traced in the original.
+
+        Args:
+            name (str, optional): Name of the branch. Defaults to "branch".
+            store (bool, optional): Whether to store the branch in the original simulation. Defaults to True.
+
+        Returns:
+            Simulation: The cloned simulation.
+        """
+        if name == self.branch_name:
+            return self
+        branch = self.clone()
+        self.branches[name] = branch
+        branch.branch_name = name
+        if self.trace:
+            branch.trace = True
+            branch.tracer = self.tracer
+        return branch
+
+    def derivative(
+        self, variable: str, wrt: str, period: Period = None, delta: float = 1
+    ) -> ArrayLike:
+        """
+        Compute the derivative of a variable w.r.t another variable.
+
+        Args:
+            variable (str): The variable to differentiate.
+            wrt (str): The variable to differentiate with respect to.
+            period (Period): The period for which to compute the derivative.
+            delta (float): The infinitesimal to use for the derivative.
+
+        Returns:
+            ArrayLike: The derivative.
+        """
+
+        if period is not None and not isinstance(period, Period):
+            period = periods.period(period)
+        elif period is None and self.default_calculation_period is not None:
+            period = periods.period(self.default_calculation_period)
+
+        alt_sim = self.clone()
+        for computed_variable in alt_sim.tax_benefit_system.variables:
+            if computed_variable not in self.input_variables:
+                alt_sim.delete_arrays(computed_variable)
+        alt_sim.set_input(wrt, period, self.calculate(wrt, period) + delta)
+        original_value = self.calculate(variable, period)
+        new_value = alt_sim.calculate(variable, period)
+        difference = new_value - original_value
+        return difference / delta
